// Sets v 2.00
// Hubert Chan, Nov 13, 1997

#include "sets.hh"

#define INTBITS sizeof(int) * 8

static int counter;
#define DOONWHOLEARRAY(x) for(counter = 0; counter < setsize; counter++) x

Set::Set(Set const &Set2): SetElements(new int[setsize]) {
  DOONWHOLEARRAY(SetElements[counter] = Set2.SetElements[counter]);
}

void Set::AddElem(SetElement Element) {
  int Segment = (Element - minelement) / INTBITS,
    Offset  = (Element - minelement) % INTBITS;

  SetElements[Segment] |= 1 << Offset;
}

void Set::RemoveElem(SetElement Element) {
  int Segment = (Element - minelement) / INTBITS,
    Offset  = (Element - minelement) % INTBITS;

  SetElements[Segment] &= ~(1 << Offset);
}

void Set::ClearSet() {
  DOONWHOLEARRAY(SetElements[counter] = 0);
}

static int bitcount(int x) {
  int n = 0;
  if(x)
    do
      n++;
    while(0 != (x = x&(x-1)));
  return(n);
}

int Set::Card() const {
  int n = 0;
  DOONWHOLEARRAY(n += bitcount(SetElements[counter]));
  return n;
}

Set Set::operator=(Set const &Set2) {
  DOONWHOLEARRAY(SetElements[counter] = Set2.SetElements[counter]);
  return *this;
}

Set Set::operator+(Set const &Set2) const {
  Set Result;
  DOONWHOLEARRAY(Result.SetElements[counter] = SetElements[counter] | Set2.SetElements[counter]);
  return Result;
}

Set Set::operator|(Set const &Set2) const {
  return (*this + Set2);
}

Set Set::operator+(SetElement Element) const {
  Set Result = *this;
  Result.AddElem(Element);
  return Result;
}

Set Set::operator|(SetElement Element) const {
  return (*this + Element);
}

Set Set::operator+=(Set const &Set2) {
  DOONWHOLEARRAY(SetElements[counter] |= Set2.SetElements[counter]);
  return *this;
}

Set Set::operator|=(Set const &Set2) {
  return (*this += Set2);
}

Set Set::operator+=(SetElement Element) {
  AddElem(Element);
  return *this;
}

Set Set::operator|=(SetElement Element) {
  return (*this += Element);
}

Set Set::operator-() {
  Set Result;
  DOONWHOLEARRAY(Result.SetElements[counter] = ~SetElements[counter]);
  return Result;
}

Set Set::operator~() {
  return -(*this);
}

Set Set::operator*(Set const &Set2) const {
  Set Result;
  DOONWHOLEARRAY(Result.SetElements[counter] = SetElements[counter] & Set2.SetElements[counter]);
  return Result;
}

Set Set::operator^(Set const &Set2) const {
  return (*this * Set2);
}

Set Set::operator&(Set const &Set2) const {
  return (*this * Set2);
}

Set Set::operator*=(Set const &Set2) {
  DOONWHOLEARRAY(SetElements[counter] &= Set2.SetElements[counter]);
  return *this;
}

Set Set::operator^=(Set const &Set2) {
  return (*this *= Set2);
}

Set Set::operator&=(Set const &Set2) {
  return (*this *= Set2);
}

Set Set::operator-(Set const &Set2) const {
  Set Result;
  DOONWHOLEARRAY(Result.SetElements[counter] = SetElements[counter] & (~Set2.SetElements[counter]));
  return Result;
}

Set Set::operator-(SetElement Element) const {
  Set Result = *this;
  Result.RemoveElem(Element);
  return Result;
}

Set Set::operator-=(Set const &Set2) {
  DOONWHOLEARRAY(SetElements[counter] &= !Set2.SetElements[counter]);
  return *this;
}

Set Set::operator-=(SetElement Element) {
  RemoveElem(Element);
  return *this;
}

int Set::operator==(Set const &Set2) const {
  int Equivalent = -1;
  DOONWHOLEARRAY(Equivalent &= (SetElements[counter] == Set2.SetElements[counter]));
  return Equivalent;
}

int Set::operator!=(Set const &Set2) const {
  int NotEquivalent = 0;
  DOONWHOLEARRAY(NotEquivalent |= (SetElements[counter] != Set2.SetElements[counter]));
  return (NotEquivalent != 0);
}

int Set::operator<=(Set const &Set2) const {
  int Subset = -1;
  DOONWHOLEARRAY(Subset &= (~SetElements[counter] | Set2.SetElements[counter]));
  return (Subset == -1);
}

int Set::operator<(Set const &Set2) const {
  int Subset = 0;
  DOONWHOLEARRAY(Subset |= (SetElements[counter] != Set2.SetElements[counter]));
  Subset = (Subset != 0 ) ? -1 : 0;
  DOONWHOLEARRAY(Subset &= (~SetElements[counter] | Set2.SetElements[counter]));
  return (Subset == -1);
}

int Set::operator>=(Set const &Set2) const {
  int Superset = -1;
  DOONWHOLEARRAY(Superset &= (SetElements[counter] | ~Set2.SetElements[counter]));
  return (Superset == -1);
}

int Set::operator>(Set const &Set2) const {
  int Superset = 0;
  DOONWHOLEARRAY(Superset |= (SetElements[counter] != Set2.SetElements[counter]));
  Superset = (Superset != 0) ? -1 : 0;
  DOONWHOLEARRAY(Superset &= (SetElements[counter] | ~Set2.SetElements[counter]));
  return (Superset == -1);
}

int operator<=(Set::SetElement Element, Set const &Set2) {
  int Segment = (Element - Set::minelement) / INTBITS,
    Offset  = (Element - Set::minelement) % INTBITS;

  return ((Set2.SetElements[Segment] & (1 << Offset)) != 0);
}

ostream & operator<<(ostream & s, Set const & Set2) {
  Set::SetElement counter;
  int NotFirst = 0;
  s << "{ ";
  for(counter = Set::minelement; counter <= Set::maxelement; counter++)
    if (counter <= Set2)
      if (NotFirst)
	s << ',' << counter;
      else {
	s << counter;
	NotFirst = !0;
      }
  s << " }";
  return s;
}

#include <stdio.h>

void PrintSet(Set const &Set1) {
  Set::SetElement counter;
  int NotFirst = 0;
  printf("{ ");
  for(counter = Set::minelement; counter <= Set::maxelement; counter++)
    if (counter <= Set1)
      if (NotFirst)
	printf(",%d",counter);
      else {
	printf("%d",counter);
	NotFirst = -1;
      }
  printf(" }\n");
}
